class PianoRoll {

  int x,y;
  float h,w;
  float cell_h,cell_w;
  int cell_padding = 1;
  int cell_h_nb = 24;
  int cell_w_nb = 32;
  int cell_color = 220;
  int cell_color_stroke = 200;
  int cell_color_over = 0;//color(0,0,255);
  int [] [] notes, tab;
  int instrument;
  boolean play = false;

  /* mouse interactions */
  int inX,inY;
  boolean inPressed,inLocked;

  PianoRoll(int _x, int _y, int _w, int _h, int _inst){
    x = _x;
    y = _y;
    w = _w;
    h = _h;
    cell_w = w / (cell_w_nb + cell_padding);
    cell_h = h / (cell_h_nb + cell_padding);
    tab = new int [cell_h_nb] [cell_w_nb];
    instrument = _inst;
  } 

  void display() {
    strokeWeight(1);
    stroke(cell_color_stroke);
    drawGrid();
  }

  void update(int _inX, int _inY, boolean _inPressed, boolean _inLocked) {
    inX=_inX;
    inY=_inY;
    inPressed = _inPressed;
    inLocked = _inLocked;
    Note n;
    float _x=0,_y=0;
    for (int i = 0; i < cell_h_nb; i++) {
      for (int j = 0; j < cell_w_nb; j++) {
        _x = x + (j*(cell_w + cell_padding));
        _y = y + (i*(cell_h + cell_padding));
        if ( inX >= _x 
          && inX <= _x+ cell_w
          && inY >= _y 
          && inY <= _y+ cell_h
          && inPressed) {
            if(inLocked) {
              n= new Note(36+abs(i-cell_h_nb),120,10);
              if(tab[i][j] == 0) midiOut.sendNote(n);
              tab[i][j] = 1;
              //track.addEvent(n, j); 
            } 
          else {
            tab[i][j] = 0;
            play=false;
            // !! view this topic http://processing.org/discourse/yabb2/YaBB.pl?num=1214392504/
            // track.remoteEvent(,j); // not working, NullPointerException is catched
          }

        } // isOver
      } //if
    } //if
  } // update()

  void drawGrid() {
    float _x=0.0, _y=0.0;
    for (int i = 0; i < cell_h_nb; i++) {
      for (int j = 0; j < cell_w_nb; j++) {
        _x = x + (j*(cell_w + cell_padding));
        _y = y + (i*(cell_h + cell_padding));
        if(tab[i][j] == 0) fill(cell_color);
        else fill(cell_color_over);
        rect(_x,_y,cell_w,cell_h);
      }
    }
  } // drawGrid()

  void reset() {
    for (int i = 0; i < cell_h_nb; i++) {
      for (int j = 0; j < cell_w_nb; j++) {
        tab[i][j] = 0;
      }
    } 
  }

  void sendMidi(int step) {
   Note n;
    for (int i = 0; i < cell_h_nb; i++) {
      if (tab[i][step] == 1) {
         //println("STEPMIDI " + step + " " + i);
         n= new Note(36+abs(i-cell_h_nb),120,25);
         midiOut.sendNote(n);
         /*
         !!! Polyphonic problem !!
         http://creativecomputing.cc/p5libs/promidi/midiout_method_sendnote.htm
         Be aware that you have to take care to send note off commands to release the notes otherwise you get midi hang ons.
         */
        //midiOut.sendNoteOff(n); // doesn't work
      }
    }
  }


}



