/*
 * PianoRollp5 - version 1.0
 * Simple musical application, a piano roll, with Arduino and Processing
 * Optional : a sensor (here a telemeter with the hand image) could be connected in pin 2 (Arduino code at the bottom)
 * Tips language : french
 * Code language : frenglish or englench, up to you ;-)
 * Author : Jérôme Abel
 * Contact : abel.jerome@free.fr
 * Date : 03/02/2010
 * Licence : GNU GPL
 * Libraries : controlP5, proMidi, serial (Arduino communication)
 * Problems : 
 *           * proMidi : removeEvent() doesn't work and noteOff not implemented to stop notes
 *           * buttons colors
 *           * frameRate (<= 20) problem between midi sequencer and sensor display
 * Problems discourses on proMidi :
 *           * http://codelab.fr/1499
 *           * http://processing.org/discourse/yabb2/YaBB.pl?num=1214392504/
 */

/*  libraries */
import controlP5.*;
import promidi.*;
import processing.serial.*;

/* declarations */

// controlP5
ControlP5 controlP5;
controlP5.Knob speedKnob;
int idButton;
int tmp_speed = 120;

// serial
Serial myPort;
String inString;
float capteur;  // "capteur" in french means sensor
boolean mainLocked = false;
float tmp_capteur;
boolean capteurNote = false;
int tmp_tmp_c =0, tmp_c =0;

// promidi
MidiIO midiIO;
MidiOut midiOut;
Sequencer sequencer;
Track track;
Song song;
boolean play;

// font
PFont fontA;

/* variables size */
int WIDTH = 800;
int HEIGHT = 600;
int BUTTON_X;
int TITLE_SIZE = 24;
int TITLE_MARGIN = 8;
int TITLE_LINE_SIZE = 4;
int NAV_MARGIN = 200;
int BUTTON_WIDTH=60;
int BUTTON_MARGIN=0;
int LINE_Y = TITLE_SIZE + 2*TITLE_MARGIN + TITLE_LINE_SIZE;
int BUTTON_Y = LINE_Y + TITLE_LINE_SIZE;
int INFO_HEIGHT = 40;

/* variables colors */
int BACKGROUND_COLOR = 255;
int BUTTON_COLOR_INIT = color(255,255,255,0);
int BUTTON_COLOR_OVER = color(0,0,255,80);
int BUTTON_COLOR_OVER2 = color(255,0,0,60);
int BUTTON_COLOR_LABEL = 0;
int TITLE_COLOR = 0;

/* buttons instruments + navigation */
String [] buttons = {
  "batterie","piano","guitare","synth","playNeg","pauseNeg","stopNeg","mainNeg"}; // 4 instruments + 4 boutons
int BUTTON_LENGTH = 4; // 4 instruments : batterie, piano, guitare, synth
PImage [] buttons_img = new PImage [buttons.length];
String [] buttons_tips = {
  "Choisissez l'instrument : ", 
  "synthétiseur",
  "Lancer la lecture de la séquence. Touche espace du clavier",
  "Arrêter la lecture de la séquence. Touche espace du clavier",
  "Arrêter la lecture et revenir au début de la séquence",
  "Activer/désactiver le capteur pour changer directement le tempo",
  "Changer le tempo",
  "Le bouton gauche de la souris active une note, le bouton droit l'efface",
  "La touche \"Suppr\" du clavier supprime la séquence"};
int [] tips_x = new int [buttons.length+1]; // + Knob Button

/* PianoRoll */
PianoRoll batterie,guitare,synth,piano; //115,34,80,4
int PIANO_Y = LINE_Y + BUTTON_WIDTH + 2*TITLE_LINE_SIZE +2;
boolean locked = false;
int PIANO_HEIGHT = HEIGHT - PIANO_Y - 5 - INFO_HEIGHT;
int PIANO_X = 40;
String [] notes = {
  "do","do#","ré","ré#","mi","fa","fa#","sol","sol#","la","la#","si"};
int NOTES_X = 0;
int NOTES_Y = PIANO_Y;
int NOTES_W = PIANO_X;
int NOTES_H = PIANO_HEIGHT;

/* Sequenceur */
int SEQ_MAX;
int step=0, tmp_step = 0;
float x_pos;
boolean stopped = false;

/* ************** SETUP *************** */
void setup() {
  size(WIDTH,HEIGHT);
  frame.setTitle("PianoRollp5"); 
  smooth();
  frameRate(20); // !! frame rate <= 20 for correct sensor data vizualisation 
  // font
  fontA = loadFont("AmericanTypewriter-24.vlw");
  // controlP5
  controlP5 = new ControlP5(this);
  // serial
  myPort = new Serial(this, "COM5", 9600);
  myPort.bufferUntil('\n'); // pour creer un serialEvent
  // pianoRoll
  piano = new PianoRoll(PIANO_X,PIANO_Y+1,WIDTH-8-PIANO_X,PIANO_HEIGHT,115);
  // proMidi
  midiIO = MidiIO.getInstance(this);
  midiOut = midiIO.getMidiOut(1,1);
  track = new Track("one", midiOut);
  track.setQuantization(Q._1_4);
  song = new Song("test", 200);
  song.addTrack(track);
  sequencer = new Sequencer();
  sequencer.setSong(song);
  sequencer.setLoopStartPoint(0);
  SEQ_MAX = 512/4 * piano.cell_w_nb;
  sequencer.setLoopEndPoint(SEQ_MAX);
  sequencer.setLoopCount(-1);
  // createButtons with controlP5
  createButtons();
}

/************ CREATE METHODS ***************/
void createButtons() {
  controlP5.Button imgButton;

  String imgButtonStr=""; // !! stupid thing : button's labels must be differents even if their ID are differents
  PImage img;

  // classic buttons
  for (int i = 0; i < buttons.length; i++) {
    img = loadImage(buttons[i] + ".png");
    buttons_img[i] = img;
    BUTTON_X = (BUTTON_WIDTH * i) + (i * BUTTON_MARGIN);
    if (i>=BUTTON_LENGTH) {
      BUTTON_X = BUTTON_X + NAV_MARGIN;
      BUTTON_COLOR_OVER = BUTTON_COLOR_OVER2;
    }
    tips_x[i] = BUTTON_X;
    imgButtonStr = imgButtonStr + " "; 
    imgButton = controlP5.addButton(imgButtonStr,0,BUTTON_X,BUTTON_Y,BUTTON_WIDTH,BUTTON_WIDTH);
    imgButton.setColorBackground(BUTTON_COLOR_INIT);
    imgButton.setColorActive(BUTTON_COLOR_OVER);
    imgButton.setId(i+1);
  }

  // Knob Buttons
  BUTTON_X = (BUTTON_WIDTH * buttons.length) + (buttons.length * BUTTON_MARGIN)  + NAV_MARGIN + 10;
  imgButtonStr = imgButtonStr + " "; 
  speedKnob = controlP5.addKnob(imgButtonStr,0,1000,tmp_speed,BUTTON_X,BUTTON_Y,BUTTON_WIDTH);
  speedKnob.setColorBackground(255);
  speedKnob.setColorActive(0);
  speedKnob.setColorLabel(0);
  speedKnob.setId(buttons.length + 1);
  tips_x[buttons.length] = BUTTON_X;
  speedKnob.setValue(tmp_speed);
}

/* ************** DRAW *************** */
void draw(){
  drawAeras();
  drawButtonsImages();
  drawPianoRoll();
  drawTips();
  drawNotes();
  drawPosition();
  drawCapteur();
}

/************ DRAW METHODS ***************/
void drawCapteur() {
  noStroke(); 
  if (mainLocked) {
    tmp_capteur = map(tmp_capteur,0,255,0,BUTTON_WIDTH);
    fill(255,0,0);
  }
  else fill(255);
  rect(680,109 - tmp_capteur,5,tmp_capteur);
}

void drawPosition() {
  if (!stopped) step = (int)(sequencer.getTickPosition() * piano.cell_w_nb / SEQ_MAX);
  else step = 0; // for Stop button, reset x position
  if (sequencer.isRunning()) {
    if (step != tmp_step) {
      piano.sendMidi(step); 
      tmp_step = step;
    }
  }
  x_pos = step * (piano.cell_w + piano.cell_padding);
  fill(255,0,0,40);
  rect(x_pos + PIANO_X, PIANO_Y, piano.cell_w,piano.h+5);
}

void drawNotes(){
  textFont(fontA, 12); 
  float y=0;
  fill(95);
  int octaveNb=2;
  for (int i = 0; i < piano.cell_h_nb ; i++) {
    y = NOTES_Y  + NOTES_H - (i * (piano.cell_h + piano.cell_padding));
    if (i==0) text(notes[i%notes.length]+"3", NOTES_W/3 , y+5);
    else if (i== 12) text(notes[i%notes.length]+"4", NOTES_W/3 , y+5);
    else text(notes[i%notes.length], NOTES_W/3 , y+5);
    strokeWeight(1);
    stroke(0);
    if(i > 0) line(0,y + 12,NOTES_W,y+7);
  }

  stroke(180);
  strokeWeight(2);
  line(NOTES_W,PIANO_Y+4+PIANO_HEIGHT/2,WIDTH,PIANO_Y+4+PIANO_HEIGHT/2);
}

void drawTips(){  
  noStroke();
  fill(150);
  rect(0,HEIGHT-INFO_HEIGHT + 1, WIDTH, INFO_HEIGHT);
  fill(255,255,0);
  textFont(fontA, 18);
  String tips="";

  // buttons tips
  for (int i = 0; i < tips_x.length; i++) {
    if ( mouseX >= tips_x[i] && mouseX <= tips_x[i] + BUTTON_WIDTH && mouseY >= BUTTON_Y && mouseY <= BUTTON_Y + BUTTON_WIDTH) {
      if (i<3) tips=buttons_tips[0] + buttons[i];
      else if (i==3) tips=buttons_tips[0] + buttons_tips[1];
      else tips=buttons_tips[i - 2];
      text("? " + tips + ".", 10, HEIGHT-14);
    }   
  }

  // piano roll tip
  textFont(fontA, 15);
  if ( mouseX >= PIANO_X && mouseX <= WIDTH && mouseY >= PIANO_Y && mouseY <= HEIGHT - INFO_HEIGHT) {
    text("? " + buttons_tips[7] + ".", 10, HEIGHT-22);
    text("? " + buttons_tips[8] + ".", 10, HEIGHT-5);
  }   
}

void drawPianoRoll(){
  piano.update(mouseX,mouseY,mousePressed,locked);
  piano.display();
}

void drawAeras() {
  background(BACKGROUND_COLOR);
  fill(TITLE_COLOR);
  textFont(fontA, TITLE_SIZE);
  text("PianoRollp5",TITLE_MARGIN,TITLE_SIZE + TITLE_MARGIN);
  rect(4* BUTTON_WIDTH + NAV_MARGIN,LINE_Y,500,BUTTON_WIDTH+ TITLE_MARGIN);

  stroke(TITLE_COLOR);
  strokeWeight(TITLE_LINE_SIZE);
  line(0,LINE_Y,width,LINE_Y);
  line(0,LINE_Y+ BUTTON_WIDTH+2*TITLE_LINE_SIZE,width,LINE_Y+ BUTTON_WIDTH+2*TITLE_LINE_SIZE);

  fill(255);
  textFont(fontA, 18);
  text("BPM \n"+tmp_speed, BUTTON_X + 2 * BUTTON_WIDTH + 4 + 10,  BUTTON_Y + BUTTON_WIDTH/2);
}

// Buttons images over controlP5 buttons
void drawButtonsImages() {
  int tmp_x=0;
  for (int i = 0; i < buttons.length-1; i++) {
    BUTTON_X = (BUTTON_WIDTH * i) + (i * BUTTON_MARGIN);
    if (i>=BUTTON_LENGTH) BUTTON_X = BUTTON_X + NAV_MARGIN;
    if (i==idButton-1  && idButton <=4) tint(0,0,255,255);
    else tint(255,255,255,255);
    image(buttons_img[i],BUTTON_X,BUTTON_Y,BUTTON_WIDTH,BUTTON_WIDTH);
    tmp_x = BUTTON_X;
  }

  // MainLocked image
  BUTTON_X = tmp_x + BUTTON_WIDTH  +  BUTTON_MARGIN;
  if (mainLocked) tint(255,0,0,255);
  else tint(255,255,255,255);
  image(buttons_img[7],BUTTON_X,BUTTON_Y,BUTTON_WIDTH,BUTTON_WIDTH);
}

/* ************** INTERACTIONS *************** */

void controlEvent(ControlEvent theEvent) { // from ControlP5 library
  idButton = theEvent.controller().id(); 

  //println(idButton);
  switch(idButton) {
  case 1: 
    midiOut.sendProgramChange(new ProgramChange(115)); // batterie midi sound
    break; 
  case 2: 
    midiOut.sendProgramChange(new ProgramChange(4)); // piano midi sound
    break; 
  case 3: 
    midiOut.sendProgramChange(new ProgramChange(34)); // guitare midi sound
    break;
  case 4: 
    midiOut.sendProgramChange(new ProgramChange(80)); // synth midi sound
    break;
  // PLAY/PAUSE/STOP buttons
  case 5:
    sequencer.start(); 
    stopped = false;
    break;
  case 6:
    sequencer.stop(); 
    stopped = false;
    break;
  case 7: 
    sequencer.stop();
    sequencer.setTickPosition(0);
    stopped = true;
    break;
  // Sensor button
  case 8: 
    mainLocked = !mainLocked;
    break;
  // Knob button for tempo in BPM  
  case(9): 
    tmp_speed = (int)(theEvent.controller().value()); 
    song.setTempo(tmp_speed);
    //song.setTempo((1000/tmp_speed) *60);
    break;
  }
}

void mousePressed(){
  if (mouseButton == RIGHT) locked=false;
  else if (mouseButton == LEFT) locked=true;
}

void keyPressed() {
  if (keyCode == 32) { // space key => play/pause actions
    play = !play;
    if (play) {
      sequencer.start();
      stopped = false;
    } 
    else {
      sequencer.stop();
      stopped = false;
    }
  }
  if (keyCode == DELETE) piano.reset(); // reset the notes
}

/* ******* IR telemeter sensor - ARDUINO ********* */
void serialEvent(Serial myPort) { // from serial library
  int pitch;
  float tmpf;
  int tmpi;
  inString = myPort.readStringUntil('\n');   // get the ASCII string:
  if (inString != null) {
    inString = trim(inString);
    capteur = float(inString);
    if (mainLocked) {
      
      // A bit complex, just for change tempo values with bigger step
      tmp_c = (int) capteur; // O-255
      tmpi = (int)tmp_c / 10; // 0 - 25
      if(   (tmp_c < ((tmpi + 1) * 10)) && (tmp_c >= (tmpi * 10)) && tmp_c != tmp_tmp_c) {
        //println(tmpi);
        tmpf = map(tmpi,0,25,20,1000);
        speedKnob.setValue((int) (tmpf));
        tmp_tmp_c = tmp_c;
      }
      pitch = (int) (map(capteur, 0,255,0,127));
      // not implemented, just for fun, add a controlP5 button to locked/unlocked capteurNote boolean
      if (capteurNote) midiOut.sendNote(new Note(pitch,120,25)); 
      tmp_capteur = capteur;
    }
  }
}

/*  *********************** ARDUINO CODE ************************** */
/*
int capteur = 0;
 void setup()
 {
   Serial.begin(9600);
 }
 void loop() {
   capteur = analogRead(2)/4; // 0-1023 => 0-255
   Serial.println(capteur); // send datas
 }
*/
